#ifndef SPACEVM_DPI_XDP_TEMPLATE_H
#define SPACEVM_DPI_XDP_TEMPLATE_H

#define XDP_C_CODE_TEMPLATE "\
#include <linux/bpf.h>\n\
#include <bpf/bpf_helpers.h>\n\
#include <bpf/bpf_endian.h>\n\
#include <xdp_helper.h>\n\
\n\
#define min(x, y) ((x) < (y) ? (x) : (y))\n\
\n\
#ifndef __packed\n\
    #define __packed __attribute__((packed))\n\
#endif\n\
\n\
struct {\n\
    __uint(type, BPF_MAP_TYPE_ARRAY);\n\
    __type(key, __u32);\n\
    __type(value, __u64);\n\
    __uint(max_entries, 3);\n\
} %s SEC(\".maps\");\n\
\n\
struct {\n\
    __uint(type, BPF_MAP_TYPE_ARRAY);\n\
    __type(key, __u32);\n\
    __type(value, __u64);\n\
    __uint(max_entries, 4);\n\
} %s SEC(\".maps\");\n\
\n\
FOREACH(RING_BUFFER_INIT, %s, %u);\n\
\n\
%s\n\
\n\
SEC(\"%s\")\n\
int xdp_spacevm_dpi_prog(struct xdp_md *ctx) {\n\
    int ret = 0;\n\
    __u32 key_total_pkts_barrier = 0;\n\
    __u32 key_total_bytes_barrier = 1;\n\
    __u32 key_stop_flag_barrier = 2;\n\
    __u32 key_total_pkts = 0;\n\
    __u32 key_match_pkts = 1;\n\
    __u32 key_match_bytes = 2;\n\
    __u32 key_up_discard_pkts = 3;\n\
    __u32 rx_queue_index = ctx->rx_queue_index;\n\
    __u64 *rx_total_pkts_barrier;\n\
    __u64 *rx_total_bytes_barrier;\n\
    __u64 *rx_stop_flag_barrier;\n\
    __u64 *rx_total_pkts;\n\
    __u64 *rx_match_pkts;\n\
    __u64 *rx_match_bytes;\n\
    __u64 *up_discard_pkts;\n\
    void *data_end = (void *) (long) ctx->data_end;\n\
    void *data = (void *) (long) ctx->data;\n\
    void *packet;\n\
\n\
    rx_stop_flag_barrier = bpf_map_lookup_elem(&%s, &key_stop_flag_barrier);\n\
    if (!rx_stop_flag_barrier || *rx_stop_flag_barrier > 0) {\n\
        return XDP_DROP;\n\
    }\n\
\n\
    rx_total_pkts = bpf_map_lookup_elem(&%s, &key_total_pkts);\n\
    if (rx_total_pkts) {\n\
        __sync_fetch_and_add(rx_total_pkts, 1);\n\
    }\n\
\n\
    if (cbpf_filter_func(data, data_end)) {\n\
        __u16 data_size = (__u16) (data_end - data);\n\
        rx_match_pkts = bpf_map_lookup_elem(&%s, &key_match_pkts);\n\
        rx_match_bytes = bpf_map_lookup_elem(&%s, &key_match_bytes);\n\
        rx_total_pkts_barrier = bpf_map_lookup_elem(&%s, &key_total_pkts_barrier);\n\
        rx_total_bytes_barrier = bpf_map_lookup_elem(&%s, &key_total_bytes_barrier);\n\
\n\
        if (rx_match_pkts && rx_match_bytes && rx_total_pkts_barrier && rx_total_bytes_barrier) {\n\
            __sync_fetch_and_add(rx_match_pkts, 1);\n\
            __sync_fetch_and_add(rx_match_bytes, data_size);\n\
\n\
            if (((*rx_total_pkts_barrier && *rx_match_pkts > *rx_total_pkts_barrier) ||\n\
                (*rx_total_bytes_barrier && *rx_match_bytes > *rx_total_bytes_barrier)) &&\n\
                rx_stop_flag_barrier) {\n\
                    __sync_fetch_and_add(rx_stop_flag_barrier, 1);\n\
                    return XDP_DROP;\n\
            }\n\
\n\
            switch (rx_queue_index) {\n\
                FOREACH(BRANCH, %s, %u);\n\
                default:\n\
                    packet = bpf_ringbuf_reserve(&%s_0, 2048, 0);\n\
                    if (!packet) {\n\
                        ret = -1;\n\
                        break;\n\
                    }\n\
                    if (bpf_probe_read_kernel(packet, 2, (void *) &data_size) ||\n\
                        bpf_probe_read_kernel(packet + 2, min(data_size, 2046), (void *) (long) ctx->data)) {\n\
                            bpf_ringbuf_discard(packet, 0);\n\
                            ret = -1;\n\
                            break;\n\
                    }\n\
                    bpf_ringbuf_submit(packet, 0);\n\
                    break;\n\
            }\n\
\n\
            if (ret) {\n\
                up_discard_pkts = bpf_map_lookup_elem(&%s, &key_up_discard_pkts);\n\
                if (up_discard_pkts) {\n\
                    __sync_fetch_and_add(up_discard_pkts, 1);\n\
                };\n\
            };\n\
        }\n\
    }\n\
    return XDP_DROP;\n\
}\n\
\n\
char _license[] SEC(\"license\") = \"GPL\";"

#endif //SPACEVM_DPI_XDP_TEMPLATE_H
